__version__ = "0.8.2"

from w3lib.url import canonicalize_url


def normalize_url(value: str) -> str:
    if "http://" in value or "https://" in value:
        return canonicalize_url(value)
    return value



from collections import OrderedDict
import time
from zoneinfo import ZoneInfo
from scrapy import Spider
from scrapy.exceptions import IgnoreRequest
from datetime import datetime, timedelta, timezone
import dateparser
# from browserforge.headers import HeaderGenerator
from scrapy.http import Request, Response
from scrapy import signals


class PreProcessingMiddleware:
    """
    Downloader Middleware to preprocess requests before forwarding.
    Handles deduplication
    """

    # def __init__(self):
    #     self.header_generator = HeaderGenerator()


    def process_request(self, request: Request, spider: Spider) -> None:
        _dt = request.meta.pop("_dt", None)
        _dt_format = request.meta.pop("_dt_format", None)
        if _dt:
            if not self.is_recent(_dt, _dt_format, request.url, spider):
                raise IgnoreRequest
        # browserforge =  request.meta.pop("browserforge", None)
        # if browserforge:
        #     headers = self.header_generator.generate()
        #     request.headers.update(headers)
        request.meta['requested_at'] = int(time.time() * 1000)
        return None

    def process_response(self, request: Request, response: Response, spider: Spider) -> Response:
        if "impersonate" in request.meta: # bug in scrapy-impersonate
            response.request.meta['responded_at'] = int(time.time() * 1000)
        else:
            request.meta['responded_at'] = int(time.time() * 1000)
        return response

    def is_recent(self, date_str: str, date_format: str, debug_info: str, spider: Spider) -> bool:
        """
        Check if the date is recent (within the last 2 days).
        """
        try:
            if not date_str:
                return True
            utc_today = datetime.now(ZoneInfo('UTC')).date()
            input_date = dateparser.parse(date_string=date_str, date_formats=[date_format] if date_format is not None else None).date()
            return input_date >= (utc_today - timedelta(days=2))
        except Exception as e:
            spider.logger.error(f"{str(e)}: {debug_info} ")
            return False



class ZenSpiderMiddleware:

    async def process_spider_output(self, response, result, spider):
        async for i in result:
            if isinstance(i, dict):
                new_dict = OrderedDict()
                for k,v in i.items():
                    new_dict[k] = v
                    if k == "published_at":
                        new_dict['scraped_at'] = int(time.time() * 1000)
                new_dict['requested_at'] = response.meta.get("requested_at")
                new_dict['responded_at'] = response.meta.get("responded_at")
                yield new_dict
            else:
                yield i

    async def process_start(self, start):
        async for item_or_request in start:
            yield item_or_request



import asyncio
from collections import defaultdict
import importlib
import json
from typing import Dict, List, Self
import grpc
import scrapy
from scrapy.crawler import Crawler
from scrapy.settings import Settings
from scrapy.spiders import Spider
from scrapy.exceptions import DropItem, NotConfigured
from datetime import datetime, timedelta, timezone
from scrapy.utils.defer import maybe_deferred_to_future
from scrapy.http.request import NO_CALLBACK
import dateparser
from scrapy import Item, signals
import websockets
import logging

logging.getLogger("websockets").setLevel(logging.WARNING)
logging.getLogger("pymongo").setLevel(logging.ERROR)

from spidermon.contrib.scrapy.pipelines import ItemValidationPipeline
from scrapy_zen import normalize_url
from scrapy_zen.databases import RedisDB



class PreProcessingPipeline(ItemValidationPipeline):
    """
    Pipeline to preprocess items before forwarding.
    Handles item validation, deduplication, filtering, and cleaning.
    """

    def __init__(
        self,
        settings: Settings,
        validation_enabled: bool,
        validators=None,
        stats=None,
    ) -> None:
        if validation_enabled:
            super().__init__(
                validators=validators,
                stats=stats,
                drop_items_with_errors=settings.getbool(
                    "SPIDERMON_VALIDATION_DROP_ITEMS_WITH_ERRORS", False
                ),
                add_errors_to_items=settings.getbool(
                    "SPIDERMON_VALIDATION_ADD_ERRORS_TO_ITEMS", False
                ),
                errors_field=settings.get(
                    "SPIDERMON_VALIDATION_ERRORS_FIELD", "_validation"
                ),
            )
        self.settings = settings
        self.validation_enabled = validation_enabled


    @classmethod
    def from_crawler(cls, crawler: Crawler) -> Self:
        for setting in RedisDB.settings:
            if not crawler.settings.get(setting):
                raise NotConfigured(f"{setting} is not set")
        spidermon_enabled = crawler.settings.getbool("SPIDERMON_ENABLED")
        if not spidermon_enabled:
            p = cls(
                settings=crawler.settings,
                validation_enabled=False,
            )
            crawler.signals.connect(p.spider_opened, signal=signals.spider_opened)
            crawler.signals.connect(p.spider_closed, signal=signals.spider_closed)
            return p

        validators = defaultdict(list)

        def set_validators(loader, schema):
            if type(schema) in (list, tuple):
                schema = {Item: schema}
            for obj, paths in schema.items():
                key = obj.__name__
                paths = paths if type(paths) in (list, tuple) else [paths]
                objects = [loader(v) for v in paths]
                validators[key].extend(objects)

        schema = crawler.settings.get("SPIDERMON_VALIDATION_SCHEMAS")
        if schema:
            set_validators(cls._load_jsonschema_validator, schema)
        else:
            crawler.spider.logger.warning("No schema defined. Validation disabled")

        p = cls(
            settings=crawler.settings,
            validation_enabled=True if validators else False,
            validators=validators,
            stats=crawler.stats,
        )
        crawler.signals.connect(p.spider_opened, signal=signals.spider_opened)
        crawler.signals.connect(p.spider_closed, signal=signals.spider_closed)
        return p

    async def spider_opened(self, spider: Spider) -> None:
        try:
            self.db = RedisDB()
            await self.db.connect(*[self.settings.get(setting) for setting in self.db.settings])
        except:
            raise NotConfigured("Failed to connect to DB")
        days = self.settings.getint("DB_EXPIRY_DAYS", 15)
        if days:
            spider.logger.warning("Expiration enabled for DB records")
            await self.db.cleanup(days)

    async def spider_closed(self, spider: Spider) -> None:
        if hasattr(self, "db"):
            await self.db.close()

    def is_recent(
        self, date_str: str, date_format: str, debug_info: str, spider: Spider
    ) -> bool:
        """
        Check if the date is recent (within the last 2 days).
        """
        try:
            if not date_str:
                return True
            utc_today = datetime.now(timezone.utc).date()
            input_date = dateparser.parse(
                date_string=date_str,
                date_formats=[date_format] if date_format is not None else None,
            ).date()
            return input_date >= (utc_today - timedelta(days=2))
        except Exception as e:
            spider.logger.error(f"{str(e)}: {debug_info} ")
            return False

    def _drop_item(self, item, errors):
        self.stats.add_dropped_item()
        raise DropItem(f"Validation failed! {errors}")

    async def process_item(self, item: Dict, spider: Spider) -> Dict:
        item = {
            k: "\n".join([" ".join(line.split()) for line in v.strip().splitlines()])
            if isinstance(v, str)
            else v
            for k, v in item.items()
        }
        if self.validation_enabled and "_skip_validation" not in item:
            try:
                item = super().process_item(item, spider)
            except DropItem as e:
                raise e

        _id = item.get("_id", None)
        if _id:
            _id = normalize_url(_id)
            id_exists = await self.db.exists(_id, spider.name)
            if id_exists:
                raise DropItem(f"Already exists [{_id}]")
            else:
                await self.db.insert(_id, spider.name)
        _dt = item.pop("_dt", None)
        _dt_format = item.pop("_dt_format", None)
        if _dt:
            if not self.is_recent(_dt, _dt_format, item.get("_id"), spider):
                raise DropItem(f"Outdated [{_dt}]")

        if not {k: v for k, v in item.items() if not k.startswith("_")}:
            raise DropItem("item has no data fields")

        # replace 'not scraped yet' & 'n/a' with empty or null in body (just to comply with gRPC feed)
        if item.get("body") in ["not scraped yet", "n/a"]:
            item["body"] = None
        if item.get("published_at") in ["n/a"]:
            item['published_at'] = None
        if self.settings.get("LOG_LEVEL","") == "INFO":
            spider.logger.info(item)
        return item


class PostProcessingPipeline:
    """
    Pipeline to postprocess items.
    Handles DB insertion.

    Attributes:
        settings (Settings): crawler settings object
    """

    def __init__(self, settings: Settings) -> None:
        self.settings = settings

    @classmethod
    def from_crawler(cls, crawler: Crawler) -> Self:
        for setting in RedisDB.settings:
            if not crawler.settings.get(setting):
                raise NotConfigured(f"{setting} is not set")
        p = cls(settings=crawler.settings)
        crawler.signals.connect(p.spider_opened, signal=signals.spider_opened)
        crawler.signals.connect(p.spider_closed, signal=signals.spider_closed)
        return p

    async def spider_opened(self, spider: Spider) -> None:
        try:
            self.db = RedisDB()
            await self.db.connect(*[self.settings.get(setting) for setting in self.db.settings])
        except:
            raise NotConfigured("Failed to connect to DB")

    async def spider_closed(self, spider: Spider) -> None:
        if hasattr(self, "db"):
            await self.db.close()

    async def process_item(self, item: Dict, spider: Spider) -> Dict:
        if not item.pop("_delivered", None):
            _id = item.get("_id", None)
            if _id:
                _id = normalize_url(_id)
                await self.db.remove(_id, spider.name)
        return item


class DiscordPipeline:
    """
    Pipeline to send items to a Discord webhook.

    Attributes:
        uri (str):
        exclude_fields (List[str]): List of fields that needs to be excluded for this pipeline
    """

    exclude_fields: List[str] = ["body"]

    def __init__(self, uri: str) -> None:
        self.uri = uri

    @classmethod
    def from_crawler(cls, crawler: Crawler) -> Self:
        settings = ["DISCORD_SERVER_URI"]
        for setting in settings:
            if not crawler.settings.get(setting):
                raise NotConfigured(f"{setting} is not set")
        return cls(uri=crawler.settings.get("DISCORD_SERVER_URI"))

    async def process_item(self, item: Dict, spider: Spider) -> Dict:
        await self._send(item, spider)
        return item

    async def _send(self, item: Dict, spider: Spider) -> None:
        try:
            _item = {
                k: v
                for k, v in item.items()
                if not k.startswith("_") and k.lower() not in self.exclude_fields
            }
            await maybe_deferred_to_future(
                spider.crawler.engine.download(
                    scrapy.Request(
                        url=self.uri,
                        method="POST",
                        body=json.dumps(
                            {
                                "embeds": [
                                    {
                                        "title": "Alert",
                                        "description": json.dumps(_item),
                                        "color": int("03b2f8", 16),
                                    }
                                ]
                            }
                        ),
                        headers={"Content-Type": "application/json"},
                        callback=NO_CALLBACK,
                        errback=lambda f: spider.logger.error((f.value)),
                    ),
                )
            )
            item["_delivered"] = True
        except Exception as e:
            spider.logger.error(f"Failed to send to Discord: {item['_id']}\n{str(e)}")


class SynopticPipeline:
    """
    Pipeline to send items to a Synoptic stream.

    Attributes:
        stream_id (str):
        api_key (str):
        exclude_fields (List[str]): List of fields that needs to be excluded for this pipeline
    """

    exclude_fields: List[str] = []

    def __init__(self, uri: str, stream_id: str, api_key: str) -> None:
        self.uri = uri
        self.stream_id = stream_id
        self.api_key = api_key

    @classmethod
    def from_crawler(cls, crawler: Crawler) -> Self:
        settings = ["SYNOPTIC_SERVER_URI", "SYNOPTIC_STREAM_ID", "SYNOPTIC_API_KEY"]
        for setting in settings:
            if not crawler.settings.get(setting):
                raise NotConfigured(f"{setting} is not set")
        return cls(
            uri=crawler.settings.get("SYNOPTIC_SERVER_URI"),
            stream_id=crawler.settings.get("SYNOPTIC_STREAM_ID"),
            api_key=crawler.settings.get("SYNOPTIC_API_KEY"),
        )

    async def process_item(self, item: Dict, spider: Spider) -> Dict:
        await self._send(item, spider)
        return item

    async def _send(self, item: Dict, spider: Spider) -> None:
        try:
            _item = {
                k: v
                for k, v in item.items()
                if not k.startswith("_") and k.lower() not in self.exclude_fields
            }
            await maybe_deferred_to_future(
                spider.crawler.engine.download(
                    scrapy.Request(
                        url=self.uri,
                        body=json.dumps(_item),
                        method="POST",
                        headers={
                            "content-type": "application/json",
                            "x-api-key": self.api_key,
                        },
                        callback=NO_CALLBACK,
                        errback=lambda f: spider.logger.error((f.value)),
                    )
                )
            )
            item["_delivered"] = True
        except Exception as e:
            spider.logger.error(f"Failed to send to Synoptic: {item['_id']}\n{str(e)}")


class TelegramPipeline:
    """
    Pipeline to send items to a Telegram channel.

    Attributes:
        uri (str):
        token (str):
        chat_id (str):
        exclude_fields (List[str]): List of fields that needs to be excluded for this pipeline
    """

    exclude_fields: List[str] = []

    def __init__(self, uri: str, token: str, chat_id: str) -> None:
        self.uri = uri
        self.token = token
        self.chat_id = chat_id

    @classmethod
    def from_crawler(cls, crawler: Crawler) -> Self:
        settings = ["TELEGRAM_SERVER_URI", "TELEGRAM_TOKEN", "TELEGRAM_CHAT_ID"]
        for setting in settings:
            if not crawler.settings.get(setting):
                raise NotConfigured(f"{setting} is not set")
        return cls(
            uri=crawler.settings.get("TELEGRAM_SERVER_URI"),
            token=crawler.settings.get("TELEGRAM_TOKEN"),
            chat_id=crawler.settings.get("TELEGRAM_CHAT_ID"),
        )

    async def process_item(self, item: Dict, spider: Spider) -> Dict:
        await self._send(item, spider)
        return item

    async def _send(self, item: Dict, spider: Spider) -> None:
        try:
            _item = {
                k: v
                for k, v in item.items()
                if not k.startswith("_") and k.lower() not in self.exclude_fields
            }
            await maybe_deferred_to_future(
                spider.crawler.engine.download(
                    scrapy.Request(
                        url=self.uri,
                        body=json.dumps(_item),
                        method="POST",
                        headers={
                            "content-type": "application/json",
                            "authorization": self.token,
                        },
                        callback=NO_CALLBACK,
                        errback=lambda f: spider.logger.error((f.value)),
                    )
                )
            )
            item["_delivered"] = True
        except Exception as e:
            spider.logger.error(f"Failed to send to Telegram: {item['_id']}\n{str(e)}")


class GRPCPipeline:
    """
    Pipeline to send items to a gRPC server.

    Attributes:
        uri (str):
        token (str):
        id (str):
        proto_module (str): dotted path to gRPC contract module
        exclude_fields (List[str]): List of fields that needs to be excluded for this pipeline
    """
    exclude_fields: List[str] = []


    def __init__(
        self, uri: str, token: str, id: str, id_headline: str, proto_module: str,
    ) -> None:
        self.uri = uri
        self.token = token
        self.id = id
        self.id_headline = id_headline
        self.feed_pb2 = importlib.import_module(f"{proto_module}.feed_pb2")
        self.feed_pb2_grpc = importlib.import_module(f"{proto_module}.feed_pb2_grpc")
        self.channel_grpc: grpc.aio.Channel = None
        self.client_grpc = None
        self.connected = asyncio.Event()
        self.t: asyncio.Task = None
        self.sem = asyncio.Semaphore(16)


    @classmethod
    def from_crawler(cls, crawler: Crawler) -> Self:
        settings = ["GRPC_SERVER_URI", "GRPC_TOKEN", "GRPC_ID", "GRPC_PROTO_MODULE"]
        for setting in settings:
            if not crawler.settings.get(setting):
                raise NotConfigured(f"{setting} is not set")
        p = cls(
            uri=crawler.settings.get("GRPC_SERVER_URI"),
            token=crawler.settings.get("GRPC_TOKEN"),
            id=crawler.settings.get("GRPC_ID"),
            id_headline=crawler.settings.get("GRPC_ID_HEADLINE"),
            proto_module=crawler.settings.get("GRPC_PROTO_MODULE"),
        )
        crawler.signals.connect(p.spider_opened, signal=signals.spider_opened)
        crawler.signals.connect(p.spider_closed, signal=signals.spider_closed)
        return p


    async def close_connection(self) -> None:
        if self.channel_grpc:
            await self.channel_grpc.close()
        self.channel_grpc = None
        self.client_grpc = None
        self.connected.clear()


    async def spider_opened(self, spider: Spider) -> None:
        self.connected.clear()
        self.t = asyncio.create_task(self.connect(spider))


    async def spider_closed(self, spider: Spider) -> None:
        if self.t and not self.t.done():
            self.t.cancel()
            try:
                await self.t
            except asyncio.CancelledError:
                pass
        await self.close_connection()


    async def connect(self, spider: Spider) -> None:
        while True:
            if self.connected.is_set():
                await asyncio.sleep(5.0)
                continue
            spider.logger.debug("connecting to gRPC server")
            try:
                self.channel_grpc = grpc.aio.secure_channel(self.uri, grpc.ssl_channel_credentials())
                self.client_grpc = self.feed_pb2_grpc.IngressServiceStub(self.channel_grpc)
                await asyncio.wait_for(self.channel_grpc.channel_ready(), timeout=10.0)
            except (Exception, asyncio.TimeoutError) as e:
                spider.logger.error(e)
                await self.close_connection()
                continue
            else:
                spider.logger.debug("connected to gRPC server")
                self.connected.set()


    async def process_item(self, item: Dict, spider: str) -> Dict:
        async with self.sem:
            await self._send(item, spider)
        return item


    async def _send(self, item: Dict, spider: Spider) -> None:
        _item = {
            k: v
            for k, v in item.items()
            if not k.startswith("_") and k.lower() not in self.exclude_fields
        }
        feed_id = self.id
        if ("body" in _item) and (_item["body"] is None) and self.id_headline:
            feed_id = self.id_headline
        feed_message = self.feed_pb2.FeedMessage(
            token=self.token,
            feedId=feed_id,
            messageId=item['_id'],
            message=json.dumps(_item),
        )
        await self.connected.wait()
        try:
            await self.client_grpc.SubmitFeedMessage(feed_message)
        except grpc.RpcError as e:
            spider.logger.error(f"Failed to send to gRPC server: {item['_id']}\n{str(e)}")
            await self.close_connection()
        else:
            item["_delivered"] = True
            spider.logger.debug(f"Sent to gRPC server [{feed_id}]: {item['_id']}")



class WSPipeline:
    """
    Pipeline to send items to a websocket server.

    Attributes:
        uri (str):
        exclude_fields (List[str]): List of fields that needs to be excluded for this pipeline
    """

    exclude_fields: List[str] = []

    def __init__(self, uri: str) -> None:
        self.uri = uri

    @classmethod
    def from_crawler(cls, crawler) -> Self:
        settings = ["WS_SERVER_URI"]
        for setting in settings:
            if not crawler.settings.get(setting):
                raise NotConfigured(f"{setting} is not set")
        p = cls(uri=crawler.settings.get("WS_SERVER_URI"))
        crawler.signals.connect(p.spider_opened, signal=signals.spider_opened)
        crawler.signals.connect(p.spider_closed, signal=signals.spider_closed)
        return p

    async def spider_opened(self, spider: Spider) -> None:
        self.client = await websockets.connect(self.uri)

    async def spider_closed(self, spider: Spider) -> None:
        await self.client.close()

    async def process_item(self, item: Dict, spider: Spider) -> Dict:
        await self._send(item, spider)
        return item

    async def _send(self, item: Dict, spider: Spider) -> None:
        _item = {
            k: v
            for k, v in item.items()
            if not k.startswith("_") and k.lower() not in self.exclude_fields
        }
        try:
            await self.client.send(json.dumps(_item))
            item["_delivered"] = True
            spider.logger.debug(f"Sent to WS server: {item['_id']}")
        except Exception as e:
            spider.logger.error(f"Failed to send to WS server: {item['_id']}\n{str(e)}")
            self.client = await websockets.connect(self.uri)


class HttpPipeline:
    """
    Pipeline to send items to a custom http webhook.

    Attributes:
        uri (str):
        token (str):
        exclude_fields (List[str]): List of fields that needs to be excluded for this pipeline
    """

    exclude_fields: List[str] = []

    def __init__(self, uri: str, token: str) -> None:
        self.uri = uri
        self.token = token

    @classmethod
    def from_crawler(cls, crawler: Crawler) -> Self:
        settings = ["HTTP_SERVER_URI", "HTTP_TOKEN"]
        for setting in settings:
            if not crawler.settings.get(setting):
                raise NotConfigured(f"{setting} is not set")
        p = cls(
            uri=crawler.settings.get("HTTP_SERVER_URI"),
            token=crawler.settings.get("HTTP_TOKEN"),
        )
        return p

    async def process_item(self, item: Dict, spider: Spider) -> Dict:
        await self._send(item, spider)
        return item

    async def _send(self, item: Dict, spider: Spider) -> None:
        try:
            _item = {
                k: v
                for k, v in item.items()
                if not k.startswith("_") and k.lower() not in self.exclude_fields
            }
            await maybe_deferred_to_future(
                spider.crawler.engine.download(
                    scrapy.Request(
                        url=self.uri,
                        body=json.dumps(_item),
                        method="POST",
                        headers={
                            "content-type": "application/json",
                            "authorization": self.token,
                        },
                        callback=NO_CALLBACK,
                        errback=lambda f: spider.logger.error((f.value)),
                    )
                )
            )
            item["_delivered"] = True
        except Exception as e:
            spider.logger.error(
                f"Failed to send to HttpWebhook: {item['_id']}\n{str(e)}"
            )



from scrapy.settings import BaseSettings
from pathlib import Path


def job_dir(settings: BaseSettings) -> str | None:
    path: str | None = settings["ZEN_JOBDIR"]
    if not path:
        return None
    if not Path(path).exists():
        Path(path).mkdir(parents=True)
    return path



from typing import Dict
from jinja2 import FileSystemLoader, Environment, Template
import importlib.resources
from spidermon import MonitorSuite, monitors
from scrapy.settings import Settings
from spidermon.contrib.actions.telegram.notifiers import SendTelegramMessageSpiderFinished
from spidermon.contrib.actions.discord.notifiers import SendDiscordMessageSpiderFinished
from spidermon.contrib.monitors.mixins import StatsMonitorMixin
from spidermon import Monitor
from spidermon.contrib.scrapy.monitors.monitors import CriticalCountMonitor, BaseScrapyMonitor,ErrorCountMonitor,UnwantedHTTPCodesMonitor
import logparser

    

@monitors.name("Downloader Exceptions monitor")
class CustomDownloaderExceptionMonitor(BaseScrapyMonitor):
    stat_name = "downloader/exception_count"
    ignored_stat_name = "downloader/exception_type_count/scrapy.exceptions.IgnoreRequest"
    threshold_setting = "SPIDERMON_MAX_DOWNLOADER_EXCEPTIONS"
    assert_type = "<="
    fail_if_stat_missing = False

    @monitors.name("Should have no downloader exception")
    def test_downloader_exception(self):
        if self.stat_name not in self.stats:
            self.skipTest(f"Unable to find '{self.stat_name}' in job stats.")
        count = self.stats.get(self.stat_name) - self.stats.get(self.ignored_stat_name, 0)
        threshold = self.crawler.settings.get(self.threshold_setting)
        msg = f"Expecting '{self.stat_name}' to be '{self.assert_type}' to '{threshold}'. Current value: '{count}'"
        self.assertTrue(count <= threshold, msg)


class CustomSendDiscordMessageSpiderFinished(SendDiscordMessageSpiderFinished):
    message_template = None

    def get_template(self, name: str) -> Template:
        template_dir = str(importlib.resources.files('scrapy_zen').joinpath('templates'))
        loader = FileSystemLoader(template_dir)
        env = Environment(loader=loader)
        return env.get_template('message.jinja')
    
    def get_template_context(self):
        logs: Dict = self.extract_errors_from_file(self.data['crawler'].settings)
        context = {
            "result": self.result,
            "data": self.data,
            "bot_name": self.data['crawler'].settings.get("BOT_NAME"),
            "monitors_passed": self.monitors_passed,
            "monitors_failed": self.monitors_failed,
            "include_ok_messages": self.include_ok_messages,
            "include_error_messages": self.include_error_messages,
        }
        if logs:
            context.update({**logs})
        context.update(self.context)
        return context
    
    def extract_errors_from_file(self, settings: Settings) -> Dict | None:
        f = settings.get("LOG_FILE")
        if not f:
            return            
        with open(f, "r") as f:
            logs = f.read()
        d = logparser.parse(logs)
        return {
            "critial_logs": "".join(d['log_categories']['critical_logs']['details']), 
            "error_logs": "".join(d['log_categories']['error_logs']['details'])
        }
    

class CustomSendTelegramMessageSpiderFinished(SendTelegramMessageSpiderFinished):
    message_template = None

    def get_template(self, name: str) -> Template:
        template_dir = str(importlib.resources.files('scrapy_zen').joinpath('templates'))
        loader = FileSystemLoader(template_dir)
        env = Environment(loader=loader)
        return env.get_template('message.jinja')
    
    def get_template_context(self):
        logs: Dict = self.extract_errors_from_file(self.data['crawler'].settings)
        context = {
            "result": self.result,
            "data": self.data,
            "bot_name": self.data['crawler'].settings.get("BOT_NAME"),
            "monitors_passed": self.monitors_passed,
            "monitors_failed": self.monitors_failed,
            "include_ok_messages": self.include_ok_messages,
            "include_error_messages": self.include_error_messages,
        }
        if logs:
            context.update({**logs})
        context.update(self.context)
        return context
    
    def extract_errors_from_file(self, settings: Settings) -> Dict | None:
        f = settings.get("LOG_FILE")
        if not f:
            return            
        with open(f, "r") as f:
            logs = f.read()
        d = logparser.parse(logs)
        return {
            "critial_logs": "".join(d['log_categories']['critical_logs']['details']), 
            "error_logs": "".join(d['log_categories']['error_logs']['details'])
        }
    

@monitors.name('Item validation')
class ItemValidationMonitor(Monitor, StatsMonitorMixin):

    @monitors.name('No item validation errors')
    def test_no_item_validation_errors(self):
        validation_errors = getattr(
            self.stats, 'spidermon/validation/fields/errors', 0
        )
        self.assertEqual(
            validation_errors,
            0,
            msg='Found validation errors in {} fields'.format(
                validation_errors)
        )


class SpiderCloseMonitorSuite(MonitorSuite):
    monitors = [
        CriticalCountMonitor,
        CustomDownloaderExceptionMonitor,
        ErrorCountMonitor,
        UnwantedHTTPCodesMonitor,
        ItemValidationMonitor,
    ]

    monitors_failed_actions = [
        CustomSendTelegramMessageSpiderFinished,
        # CustomSendDiscordMessageSpiderFinished,
    ]




import logging
import os
from typing import List, Self, Dict, Any
from scrapy import logformatter
from scrapy.crawler import Crawler
from scrapy.exceptions import DropItem
from scrapy import Spider
from scrapy.http import Response
from scrapy.utils.python import global_object_name
from twisted.python.failure import Failure
from scrapy.logformatter import LogFormatterResult



class ZenLogFormatter(logformatter.LogFormatter):
    YELLOW = "\033[33m"
    RED = "\033[31m"
    RESET = "\033[0m"

    def __init__(self, truncate_fields: List[str]) -> None:
        self.truncate_fields = truncate_fields

    @classmethod
    def from_crawler(cls, crawler: Crawler) -> Self:
        return cls(
            truncate_fields=crawler.settings.getlist("FORMATTER_TRUNCATE_FIELDS", []),
        )

    @staticmethod
    def truncate(value: Any, length=50) -> Any:
        if isinstance(value, str):
            return value[:length] + '...' if len(value) > length else value
        return value

    def dropped(self, item: Dict, exception: DropItem, response: Response, spider: Spider) -> LogFormatterResult:
        return {
            'level': logging.DEBUG,
            'msg': self.YELLOW + "Dropped: %(exception)s" + self.RESET + os.linesep + "%(item)s",
            'args': {
                'exception': exception,
                'item': {k:self.truncate(v) if k in self.truncate_fields else v for k,v in item.items()},
            }
        }

    def item_error(self, item: Dict, exception: DropItem, response: Response, spider: Spider) -> LogFormatterResult:
        return {
            'level': logging.ERROR,
            'msg': self.RED + "Error processing %(item)s" + self.RESET,
            'args': {
                'exception': exception,
                'item': {k:self.truncate(v) if k in self.truncate_fields else v for k,v in item.items()},
            }
        }

    def scraped(self, item: Dict, response: Response, spider: Spider) -> LogFormatterResult:
        src: Any
        if response is None:
            src = f"{global_object_name(spider.__class__)}.start_requests"
        elif isinstance(response, Failure):
            src = response.getErrorMessage()
        else:
            src = response
        return {
            "level": logging.DEBUG,
            "msg": "Scraped from %(src)s" + os.linesep + "%(item)s",
            "args": {
                "src": src,
                "item": {k:self.truncate(v) if k in self.truncate_fields else v for k,v in item.items()},
            },
        }



from scrapy.crawler import Crawler
from scrapy.statscollectors import StatsCollector
from typing import Self
import logging
from scrapy import Spider, signals
from scrapy.exceptions import NotConfigured
from scrapy.http import Request, Response
from scrapy.core.downloader import Slot
from twisted.internet import task


logger = logging.getLogger(__name__)



class ZenExtension:
    """
    Allows to calculate average latency across requests and shows logstats
    """

    def __init__(self, stats: StatsCollector) -> None:
        self.stats = stats
        self.interval: float = 60.0
        self.multiplier: float = 60.0 / self.interval
        self.task: task.LoopingCall | None = None

    @classmethod
    def from_crawler(cls, crawler: Crawler) -> Self:
        ext = cls(crawler.stats)
        crawler.signals.connect(ext.spider_opened, signal=signals.spider_opened)
        crawler.signals.connect(ext.spider_closed, signal=signals.spider_closed)
        return ext

    def spider_opened(self, spider: Spider) -> None:
        self.pagesprev: int = 0
        self.itemsprev: int = 0
        self.task = task.LoopingCall(self.log, spider)
        self.task.start(self.interval)

    def log(self, spider: Spider) -> None:
        self.calculate_stats()
        msg = (
            "Crawled %(pages)d pages (at %(pagerate)d pages/min), "
            "scraped %(items)d items (at %(itemrate)d items/min) [%(spider_name)s]"
        )
        log_args = {
            "pages": self.pages,
            "pagerate": self.prate,
            "items": self.items,
            "itemrate": self.irate,
            "spider_name": spider.name,
        }
        logger.info(msg, log_args, extra={"spider": spider})

    def calculate_stats(self) -> None:
        self.items: int = self.stats.get_value("item_scraped_count", 0)
        self.pages: int = self.stats.get_value("response_received_count", 0)
        self.irate: float = (self.items - self.itemsprev) * self.multiplier
        self.prate: float = (self.pages - self.pagesprev) * self.multiplier
        self.pagesprev, self.itemsprev = self.pages, self.items

    def spider_closed(self, spider: Spider, reason: str) -> None:
        if self.task and self.task.running:
            self.task.stop()

        rpm_final, ipm_final = self.calculate_final_stats(spider)
        self.stats.set_value("responses_per_minute", rpm_final)
        self.stats.set_value("items_per_minute", ipm_final)

    def calculate_final_stats(
        self, spider: Spider
    ) -> tuple[None, None] | tuple[float, float]:
        start_time = self.stats.get_value("start_time")
        finish_time = self.stats.get_value("finish_time")

        if not start_time or not finish_time:
            return None, None

        mins_elapsed = (finish_time - start_time).seconds / 60

        if mins_elapsed == 0:
            return None, None

        items = self.stats.get_value("item_scraped_count", 0)
        pages = self.stats.get_value("response_received_count", 0)

        return (pages / mins_elapsed), (items / mins_elapsed)


class ZenAutoThrottle:
    def __init__(self, crawler: Crawler):
        self.crawler: Crawler = crawler
        if not crawler.settings.getbool("ZEN_AUTOTHROTTLE_ENABLED"):
            raise NotConfigured

        self.inc_factor: float = crawler.settings.getfloat(
            "ZEN_AUTOTHROTTLE_BACKOFF", 1.5
        )
        self.debug: bool = crawler.settings.getbool("ZEN_AUTOTHROTTLE_DEBUG")
        self.target_concurrency: float = crawler.settings.getfloat(
            "ZEN_AUTOTHROTTLE_TARGET_CONCURRENCY"
        )
        if self.target_concurrency <= 0.0:
            raise NotConfigured(
                f"ZEN_AUTOTHROTTLE_TARGET_CONCURRENCY "
                f"({self.target_concurrency!r}) must be higher than 0."
            )
        crawler.signals.connect(self._spider_opened, signal=signals.spider_opened)
        crawler.signals.connect(
            self._response_downloaded, signal=signals.response_downloaded
        )

    @classmethod
    def from_crawler(cls, crawler: Crawler) -> Self:
        return cls(crawler)

    def _spider_opened(self, spider: Spider) -> None:
        self.mindelay = self._min_delay(spider)
        self.maxdelay = self._max_delay(spider)
        spider.download_delay = self._start_delay(spider)  # type: ignore[attr-defined]

    def _min_delay(self, spider: Spider) -> float:
        s = self.crawler.settings
        return getattr(spider, "download_delay", s.getfloat("DOWNLOAD_DELAY"))

    def _max_delay(self, spider: Spider) -> float:
        return self.crawler.settings.getfloat("ZEN_AUTOTHROTTLE_MAX_DELAY")

    def _start_delay(self, spider: Spider) -> float:
        return max(
            self.mindelay, self.crawler.settings.getfloat("ZEN_AUTOTHROTTLE_START_DELAY")
        )

    def _response_downloaded(
        self, response: Response, request: Request, spider: Spider
    ) -> None:
        key, slot = self._get_slot(request, spider)
        latency = request.meta.get("download_latency")
        if (
            latency is None
            or slot is None
            or request.meta.get("autothrottle_dont_adjust_delay", False) is True
        ):
            return

        olddelay = slot.delay
        self._adjust_delay(slot, latency, response)
        if self.debug:
            diff = slot.delay - olddelay
            size = len(response.body)
            conc = len(slot.transferring)
            logger.info(
                "slot: %(slot)s | conc:%(concurrency)2d | "
                "delay:%(delay)5d ms (%(delaydiff)+d) | "
                "latency:%(latency)5d ms | size:%(size)6d bytes",
                {
                    "slot": key,
                    "concurrency": conc,
                    "delay": slot.delay * 1000,
                    "delaydiff": diff * 1000,
                    "latency": latency * 1000,
                    "size": size,
                },
                extra={"spider": spider},
            )

    def _get_slot(
        self, request: Request, spider: Spider
    ) -> tuple[str | None, Slot | None]:
        key: str | None = request.meta.get("download_slot")
        if key is None:
            return None, None
        assert self.crawler.engine
        return key, self.crawler.engine.downloader.slots.get(key)

    def _adjust_delay(self, slot: Slot, latency: float, response: Response) -> None:
        """Define delay adjustment policy"""

        # If a server needs `latency` seconds to respond then
        # we should send a request each `latency/N` seconds
        # to have N requests processed in parallel
        target_delay = latency / self.target_concurrency
        current_delay = slot.delay

        if response.status == 429:
            new_delay = current_delay * self.inc_factor
        else:
            # Adjust the delay to make it closer to target_delay
            new_delay = (slot.delay + target_delay) / 2.0

            # If target delay is bigger than old delay, then use it instead of mean.
            # It works better with problematic sites.
            new_delay = max(target_delay, new_delay)

            # Make sure self.mindelay <= new_delay <= self.max_delay
            new_delay = min(max(self.mindelay, new_delay), self.maxdelay)

            # Dont adjust delay if response status != 200 and new delay is smaller
            # than old one, as error pages (and redirections) are usually small and
            # so tend to reduce latency, thus provoking a positive feedback by
            # reducing delay instead of increase.
            if response.status != 200 and new_delay <= slot.delay:
                return

        slot.delay = new_delay



from scrapy.settings import Settings
import os
import importlib.resources
from pathlib import Path
from scrapy.crawler import Crawler

from .logformatter import ZenLogFormatter


class SpidermonAddon:
    def update_settings(self, settings: Settings) -> None:
        settings.set("SPIDERMON_ENABLED", True, "addon")
        settings["EXTENSIONS"].update(
            {
                "spidermon.contrib.scrapy.extensions.Spidermon": 500,
            }
        )
        settings.set(
            "SPIDERMON_SPIDER_CLOSE_MONITORS",
            {"scrapy_zen.monitors.SpiderCloseMonitorSuite": 543},
            "addon",
        )
        settings.set("SPIDERMON_MAX_CRITICALS", 0, "addon")
        settings.set("SPIDERMON_MAX_DOWNLOADER_EXCEPTIONS", 0, "addon")
        settings.set("SPIDERMON_MAX_ERRORS", 0, "addon")
        settings.set(
            "SPIDERMON_UNWANTED_HTTP_CODES",
            {
                401: 0,
                403: 0,
                429: 0,
                400: 0,
            },
            "addon",
        )
        # item validation
        validation_schema = settings.get("ZEN_VALIDATION_SCHEMA")
        if validation_schema:
            if validation_schema == "NEWS":
                settings.set(
                    "SPIDERMON_VALIDATION_SCHEMAS",
                    [
                        str(importlib.resources.files("scrapy_zen").joinpath("schemas/news.json"))
                    ],
                    "addon",
                )
            else:
                settings.set(
                    "SPIDERMON_VALIDATION_SCHEMAS", [validation_schema], "addon"
                )
            settings.set("SPIDERMON_VALIDATION_ADD_ERRORS_TO_ITEMS", True, "addon")
            settings.set("SPIDERMON_VALIDATION_DROP_ITEMS_WITH_ERRORS", True, "addon")
        # telegram (disabled)
        settings.set(
            "SPIDERMON_TELEGRAM_SENDER_TOKEN",
            os.getenv("SPIDERMON_TELEGRAM_SENDER_TOKEN"),
            "addon",
        )
        settings.set("SPIDERMON_TELEGRAM_RECIPIENTS", ["-1002462968579"], "addon")
        settings.set(
            "SPIDERMON_TELEGRAM_NOTIFIER_INCLUDE_ERROR_MESSAGES", True, "addon"
        )
        # discord
        settings.set(
            "SPIDERMON_DISCORD_WEBHOOK_URL",
            os.getenv("SPIDERMON_DISCORD_WEBHOOK_URL"),
            "addon",
        )


class ZenAddon:

    def __init__(self, crawler: Crawler) -> None:
        super().__init__()
        self.crawler = crawler

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler)

    def update_settings(self, settings: Settings) -> None:
        settings.set("LOG_FORMAT", "%(asctime)s.%(msecs)03d [%(name)s] %(levelname)s: %(message)s", "addon")
        settings.set("TELNETCONSOLE_PASSWORD", "toor", "addon")
        # logger
        settings.set("LOG_FORMATTER", ZenLogFormatter, "addon")

        # Database
        settings.set("DB_NAME", os.getenv("DB_NAME"), "addon")
        settings.set("DB_USER", os.getenv("DB_USER"), "addon")
        settings.set("DB_PASS", os.getenv("DB_PASS"), "addon")
        settings.set("DB_HOST", os.getenv("DB_HOST"), "addon")
        settings.set("DB_PORT", os.getenv("DB_PORT"), "addon")

        # discord
        settings.set("DISCORD_SERVER_URI", os.getenv("DISCORD_SERVER_URI"), "addon")

        # synoptic
        settings.set("SYNOPTIC_SERVER_URI", os.getenv("SYNOPTIC_SERVER_URI"), "addon")
        settings.set("SYNOPTIC_STREAM_ID", os.getenv("SYNOPTIC_STREAM_ID"), "addon")
        settings.set("SYNOPTIC_API_KEY", os.getenv("SYNOPTIC_API_KEY"), "addon")

        # telegram
        settings.set("TELEGRAM_SERVER_URI", os.getenv("TELEGRAM_SERVER_URI"), "addon")
        settings.set("TELEGRAM_TOKEN", os.getenv("TELEGRAM_TOKEN"), "addon")
        settings.set("TELEGRAM_CHAT_ID", os.getenv("TELEGRAM_CHAT_ID"), "addon")

        # gRPC
        settings.set("GRPC_SERVER_URI", os.getenv("GRPC_SERVER_URI"), "addon")
        settings.set("GRPC_TOKEN", os.getenv("GRPC_TOKEN"), "addon")
        settings.set("GRPC_ID", os.getenv("GRPC_ID"), "addon")
        settings.set("GRPC_ID_HEADLINE", os.getenv("GRPC_ID_HEADLINE"), "addon")
        settings.set("GRPC_PROTO_MODULE", os.getenv("GRPC_PROTO_MODULE"), "addon")

        # websocket
        settings.set("WS_SERVER_URI", os.getenv("WS_SERVER_URI"), "addon")

        # custom http webhook
        settings.set("HTTP_SERVER_URI", os.getenv("HTTP_SERVER_URI"), "addon")
        settings.set("HTTP_TOKEN", os.getenv("HTTP_TOKEN"), "addon")

        # scrapy-zyte-api
        settings.set("ZYTE_API_KEY", os.getenv("ZYTE_API_KEY"), "addon")
        settings["DOWNLOADER_MIDDLEWARES"].update(
            {"scrapy_zyte_api.ScrapyZyteAPIDownloaderMiddleware": 643}
        )
        settings["SPIDER_MIDDLEWARES"].update(
            {
                "scrapy_zen.middlewares.ZenSpiderMiddleware": 10,
                "scrapy_zyte_api.ScrapyZyteAPISpiderMiddleware": 100,
                "scrapy_zyte_api.ScrapyZyteAPIRefererSpiderMiddleware": 1000,
            }
        )
        settings.set(
            "REQUEST_FINGERPRINTER_CLASS",
            "scrapy_zyte_api.ScrapyZyteAPIRequestFingerprinter",
            "addon",
        )

        # scrapy-playwright
        settings.set(
            "PLAYWRIGHT_ABORT_REQUEST",
            lambda req: req.resource_type == "image" or ".jpg" in req.url,
            "addon",
        )
        settings.set("PLAYWRIGHT_PROCESS_REQUEST_HEADERS", None, "addon")

        # download handler
        settings.set("DOWNLOAD_HANDLERS", {
            "http": "scrapy_zen.handler.ZenDownloadHandler",
            "https": "scrapy_zen.handler.ZenDownloadHandler",
        }, "addon")

        settings.set("USER_AGENT", None, "addon")
        settings.set("COOKIES_ENABLED", False, "addon")
        settings.set(
            "TWISTED_REACTOR",
            "twisted.internet.asyncioreactor.AsyncioSelectorReactor",
            "addon",
        )

        # dupefilter
        jobdir = str(Path(importlib.resources.files("scrapy_zen").joinpath("jobdirs")) / self.crawler.spider.name)
        settings.set("ZEN_JOBDIR", jobdir, "addon")
        settings.set(
            "DUPEFILTER_CLASS",
            "scrapy_zen.dupefilters.ZenDupeFilter",
            "addon",
        )

        # extensions
        settings["EXTENSIONS"].update(
            {
                "scrapy_zen.extensions.ZenAutoThrottle": 551,
                "scrapy_zen.extensions.ZenExtension": 552,
                "scrapy.extensions.logstats.LogStats": None, # disable default logstats (ZenExtension will handle it)
            }
        )



import random
from scrapy_playwright.handler import ScrapyPlaywrightDownloadHandler
from scrapy_zyte_api import ScrapyZyteAPIDownloadHandler
from scrapy_girgit import ImpersonateDownloadHandler
from scrapy.core.downloader.handlers.http import HTTPDownloadHandler
from twisted.internet.defer import Deferred
from scrapy.http import Request, Response
from scrapy import Spider
from scrapy.crawler import Crawler
from typing import Self
from scrapy import signals


impersonate_fingerprints = [
    "chrome99",
    "chrome99_android",
    "chrome100",
    "chrome101",
    "chrome104",
    "chrome107",
    "chrome110",
    "chrome116",
    "chrome119",
    "chrome120",
    "chrome123",
    "chrome124",
    "chrome131",
    "chrome131_android",
    "chrome133a",
    "edge99",
    "edge101",
    "safari15_3",
    "safari15_5",
    "firefox133",
    "firefox135",
]



class ZenDownloadHandler:
    lazy = False

    def __init__(self, crawler: Crawler):
        self.playwright_handler = ScrapyPlaywrightDownloadHandler.from_crawler(crawler)
        self.zyte_handler = ScrapyZyteAPIDownloadHandler.from_crawler(crawler)
        self.impersonate_handler = ImpersonateDownloadHandler.from_crawler(crawler)
        self.scrapy_default_handler = HTTPDownloadHandler.from_crawler(crawler)

    @classmethod
    def from_crawler(cls, crawler: Crawler) -> Self:
        h = cls(crawler)
        crawler.signals.connect(h.spider_closed, signal=signals.spider_closed)
        return h

    def download_request(self, request: Request, spider: Spider) -> Deferred | Deferred[Response]:
        if request.meta.get('playwright'):
            return self.playwright_handler.download_request(request, spider)

        elif request.meta.get("impersonate"):
            impersonate = request.meta.get("impersonate")
            if impersonate == "random":
                request.meta["impersonate"] = random.choice(impersonate_fingerprints)
            return self.impersonate_handler.download_request(request, spider)

        elif request.meta.get('zyte_api_automap'):
            return self.zyte_handler.download_request(request, spider)
        else:
            return self.scrapy_default_handler.download_request(request, spider)

    async def spider_closed(self, spider: Spider) -> None:
        self.impersonate_handler.close()
        self.scrapy_default_handler.close()
        self.playwright_handler.close()
        self.zyte_handler.close()



from abc import ABC, abstractmethod
import time
from typing import List
import redis.asyncio as redis



class DB(ABC):

    @abstractmethod
    async def connect(self, host: str, port: int) -> None:
        ...

    @abstractmethod
    async def insert(self, id: str, spider_name: str) -> None:
        ...

    @abstractmethod
    async def exists(self, id: str, spider_name: str) -> bool:
        ...

    @abstractmethod
    async def remove(self, id: str, spider_name: str) -> None:
        ...

    @abstractmethod
    async def cleanup(self, days: int) -> None:
        ...

    @abstractmethod
    async def close(self) -> None:
        ...


class RedisDB(DB):
    settings: List[str] = ["DB_HOST", "DB_PORT", "DB_PASS"]
    PROCESSED_IDS_ZSET: str = "processed:ids:zset"

    async def connect(self, host: str = "localhost", port: int = 6379, password: str = None) -> None:
        self.r = redis.Redis(
            host=host,
            port=port,
            password=password,
            socket_timeout=5,
            socket_connect_timeout=5,
            decode_responses=True,
        )
        await self.r.ping()

    async def insert(self, id: str, spider_name: str) -> None:
        unique_id = f"{spider_name}_{id}"
        timestamp = int(time.time())
        await self.r.zadd(self.PROCESSED_IDS_ZSET, {unique_id: timestamp})

    async def exists(self, id: str, spider_name: str) -> bool:
        unique_id = f"{spider_name}_{id}"
        score = await self.r.zscore(self.PROCESSED_IDS_ZSET, unique_id)
        return score is not None

    async def remove(self, id: str, spider_name: str) -> None:
        unique_id = f"{spider_name}_{id}"
        self.r.delete(unique_id)

    async def cleanup(self, days: int) -> None:
        threshold_timestamp = int(time.time()) - (days * 24 * 60 * 60)
        await self.r.zremrangebyscore(self.PROCESSED_IDS_ZSET, 0, threshold_timestamp)

    async def close(self) -> None:
        if hasattr(self, "r"):
            await self.r.aclose()



from typing import Any
from scrapy.contracts import Contract
from scrapy.exceptions import ContractFail
from itemadapter import ItemAdapter, is_item


class ScrapesContract(Contract):
    """Contract to check presence of fields in scraped items
    @scrapes page_name page_body
    """

    name = "scrapes"

    def post_process(self, output: list[Any]) -> None:
        for x in output:
            if is_item(x):
                item = ItemAdapter(x)
                missing = [arg for arg in self.args if arg not in item]
                if missing:
                    missing_fields = ", ".join(missing)
                    raise ContractFail(f"Missing fields: {missing_fields}")
                for arg in self.args:
                    if not item.get(arg):
                        raise ContractFail(f"Missing value for field: {arg}")